/*
 * Written for SGI DMedia API by Masanao Izumo <mo@goice.co.jp>
 * Ported to OSS by David Olofson <do@reologica.se>, 2001
 *
FIXME: The command FIFO I added isn't timestamped.
FIXME: /David
 *
 * TODO:
 *	* Add resampling filter to avoid sound card
 *	  sample rate dependency.
 *
 *	* Add compressor to avoid clipping distortion.
 *
 *	* Add mmap() support for lowlatency sfx.
 */

#define DBG	if(0)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include "audio.h"
#include "audiodev.h"
#include "sfifo.h"

#define BUFSIZE	32
#define BUFFERS		3
#define	MAX_COMMANDS	64


/*----------------------------------------------------------
	Asynchronous command interface stuff
----------------------------------------------------------*/

typedef struct
{
	enum
	{
		CMD_VOLUME = 0,
		CMD_STOP,
		CMD_PLAY,
		CMD_PLAY_LOOPED
	} action;
	int	cid;
	union
	{
		int	vol;
		struct
		{
			char	*start;
			int	length;
		} sample;
	} args;
} command_t;

static sfifo_t commands;


/*----------------------------------------------------------
	Audio channel
----------------------------------------------------------*/

typedef signed char audio_samp_t;
typedef signed short out_samp_t;

typedef struct
{
	audio_samp_t *samp;
	int nsamp;	/* number of samples */
	int position;
	int vol;	/* percent of default */
	int looping;	/* restart when end is reached */
} audio_channel_t;

static audio_channel_t audio_channel[MAX_CHANNEL];


/*----------------------------------------------------------
	Engine stuff
----------------------------------------------------------*/

static audiodev_t adev;
static pthread_t engine_thread;

static int running = 0;
static int master_vol = 100<<2;


/*----------------------------------------------------------
	Engine code
----------------------------------------------------------*/

static void __play_stop(int cid)
{
	audio_channel_t *c;
	c = audio_channel + cid;
	c->samp = NULL;
}


static void __play(int cid, char *samp, int nsamp, int loop)
{
	audio_channel_t *c;
	c = audio_channel + cid;
	c->samp = (audio_samp_t *) samp;
	c->nsamp = nsamp;
	c->position = 0;
	c->looping = loop;
}


static void __audio_volume(int cid, int vol)
{
	if (cid == -1)
	{
		int i;
		for (i = 0; i < MAX_CHANNEL; i++)
			audio_channel[i].vol = vol;
	}
	else
		audio_channel[cid].vol = vol;
}


static void __run_commands(void)
{
	while(sfifo_used(&commands) >= sizeof(command_t))
	{
		command_t cmd;
		if(sfifo_read(&commands, &cmd, sizeof(cmd)) < 0)
		{
			fprintf(stderr, "audio.c: Engine failure!\n");
			running = 0;
			return;
		}
		switch(cmd.action)
		{
		  case CMD_VOLUME:
			DBG printf("Got CMD_VOLUME.\n");
			__audio_volume(cmd.cid, cmd.args.vol);
			break;
		  case CMD_STOP:
			DBG printf("Got CMD_STOP.\n");
			__play_stop(cmd.cid);
			break;
		  case CMD_PLAY:
			DBG printf("Got CMD_PLAY.\n");
			__play(cmd.cid, cmd.args.sample.start,
					cmd.args.sample.length, 0);
			break;
		  case CMD_PLAY_LOOPED:
			DBG printf("Got CMD_PLAY_LOOPED.\n");
			__play(cmd.cid, cmd.args.sample.start,
					cmd.args.sample.length, 1);
			break;
		}

	}
}


static void __push_command(command_t *cmd)
{
	if(!running)
		return;

	if(sfifo_space(&commands) < sizeof(command_t))
	{
		fprintf(stderr, "WARNING: Audio command FIFO overflow!\n");
		return;
	}
	sfifo_write(&commands, cmd, sizeof(command_t));
}


static void __mix(int *buf, int frames)
{
	int i;
	memset(buf, 0, frames * sizeof(int));
	for (i = 0; i < MAX_CHANNEL; i++)
	{
		audio_channel_t *c;
		c = audio_channel + i;

		if(!c->samp)
			continue;
		
		/* Mix from 8 bit source */
		if(c->position < c->nsamp)
		{
			int j, n, v;
			audio_samp_t *s;

			n = c->nsamp - c->position;
			v = c->vol;
			if (n > frames)
				n = frames;
			s = c->samp + c->position;

			if (v == DEFAULT_VOL)
			{
				for (j = 0; j < n; j++)
					buf[j] += s[j]<<8;
			}
			else
			{
				int vm = (double)v / (double)DEFAULT_VOL;
				for (j = 0; j < n; j++)
					buf[j] += s[j] * vm;
			}
			c->position += n;
		}

		/* Handle end-of-samples */
		if(c->position >= c->nsamp)
		{
			if(c->looping)
				c->position = 0;
			else
				c->samp = NULL;
		}
	}
}


/*
 * Apply master volume and convert with saturation
 */
static void __clip(int *inbuf, out_samp_t *outbuf, int frames)
{
	int i;
	for(i = 0; i < frames; ++i)
	{
		int s = (int)inbuf[i];
		s *= master_vol;
		s >>= 8;
		if(s < -32768)
			outbuf[i] = -32768;
		else if(s > 32767)
			outbuf[i] = 32767;
		else
			outbuf[i] = s;
	}
}


void *__audio_engine(void *dummy)
{
	int mixbuf[BUFSIZE];
	out_samp_t outbuf[BUFSIZE];
	while(running)
	{
		__run_commands();
		__mix(mixbuf, BUFSIZE);
		__clip(mixbuf, outbuf, BUFSIZE);
		write(adev.outfd, outbuf, sizeof(outbuf));
	}
	audiodev_close(&adev);
	return NULL;
}


/*----------------------------------------------------------
	Open/close code
----------------------------------------------------------*/

int audio_open(void)
{
	int i;
	running = 0;

	if(sfifo_init(&commands, sizeof(command_t) * MAX_COMMANDS) < 0)
	{
		fprintf(stderr, "audio.c: Failed to set up audio engine!\n");
		return -1;
	}

	memset(audio_channel, 0, sizeof(audio_channel));

	for (i = 0; i < MAX_CHANNEL; i++)
		audio_channel[i].vol = DEFAULT_VOL;

	audiodev_init(&adev);
	adev.rate = 8000;
	adev.fragmentsize = BUFSIZE * sizeof(short) * 2;
	adev.fragments = BUFFERS;
	running = 1;
	if(audiodev_open(&adev) < 0)
	{
		sfifo_close(&commands);
		fprintf(stderr, "audio.c: Failed to open audio!\n");
		running = 0;
		return -2;
	}

	if(pthread_create(&engine_thread, NULL, __audio_engine, NULL))
	{
		fprintf(stderr, "audio.c: Failed to start audio engine!\n");
		return -3;
	}

	return 0;
}


void audio_close(void)
{
	if(!running)
		return;

	running = 0;
	printf("Waiting for audio engine to stop...\n");
	pthread_join(engine_thread, NULL);
	printf("  Ok.\n");
}


/*----------------------------------------------------------
	Channel control API code
----------------------------------------------------------*/

void audio_master_volume(int vol)
{
	master_vol = vol<<2;
}


void audio_volume(int cid, int vol)
{
	command_t cmd;
	cmd.action = CMD_VOLUME;
	cmd.cid = cid;
	cmd.args.vol = vol;
	__push_command(&cmd);
}

void audio_play(int cid, char* samp, int nsamp)
{
	command_t cmd;
	cmd.action = CMD_PLAY;
	cmd.cid = cid;
	cmd.args.sample.start = samp;
	cmd.args.sample.length = nsamp;
	__push_command(&cmd);
}

void audio_play_looped(int cid, char* samp, int nsamp)
{
	command_t cmd;
	cmd.action = CMD_PLAY_LOOPED;
	cmd.cid = cid;
	cmd.args.sample.start = samp;
	cmd.args.sample.length = nsamp;
	__push_command(&cmd);
}

void audio_stop(int cid)
{
	command_t cmd;
	if(cid < 0)
		return;

	cmd.action = CMD_STOP;
	cmd.cid = cid;
	__push_command(&cmd);
}


/*
 * audio_is_active() bypasses the lock-free FIFO.
 * OTOH, it's meant to be asynchronous (I think...),
 * so it should do the right thing, provided the
 * reads are atomic.
 *		/David
 */
int audio_is_active(int cid)
{
	audio_channel_t *c;
	if(!running)
		return 0;

	c = audio_channel + cid;
	return c->samp != NULL;
}
